/**
 * Copyright (C) 2023 Chris Noxz
 * Author(s): Chris Noxz <chris@noxz.tech>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include "browser.h"
#include "config.h"


struct Client*
client_create(const gchar *uri, WebKitWebView *related_wv, gboolean show,
              gboolean focus_tab)
{
	struct Client          *c;                  /* client to be */
	gchar                  *u;                  /* translated uri */
	GtkWidget              *e,                  /* event box */
	                       *t;                  /* tab box */

	/* communicate uri through ipc if constructed */
	if (uri != NULL && gl.ipc == IPC_CLIENT) {
		if (ipc_send((u = resolve_uri(uri))) <= 0 || ipc_send("\n") <= 0)
			fprintf(stderr, __NAME__": Could not send message '%s'\n", u);
		g_free(u);
		return NULL; /* the request was handled elsewhere */
	}

	if (!(c = calloc(1, sizeof(struct Client))))
		die("%s: fatal: calloc failed\n", __NAME__);

	if (related_wv == NULL)
		c->wv = webkit_web_view_new();
	else
		c->wv = webkit_web_view_new_with_related_view(related_wv);

	/* connect signal callbacks */
	CB(c->wv, "button-release-event",           cb_wv_hid, c);
	CB(c->wv, "close",                          cb_wv_close, c);
	CB(c->wv, "context-menu",                   cb_context_menu, c);
	CB(c->wv, "create",                         cb_wv_create, NULL);
	CB(c->wv, "decide-policy",                  cb_wv_decide_policy, NULL);
	CB(c->wv, "key-press-event",                cb_wv_hid, c);
	CB(c->wv, "load-changed",                   cb_wv_load_changed, c);
	CB(c->wv, "load-failed-with-tls-errors",    cb_wv_tls_load_failed, c);
	CB(c->wv, "mouse-target-changed",           cb_wv_hover, c);
	CB(c->wv, "notify::estimated-load-progress",cb_wv_load_progress_changed, c);
	CB(c->wv, "notify::favicon",                cb_favicon_changed, c);
	CB(c->wv, "notify::title",                  cb_title_changed, c);
	CB(c->wv, "notify::uri",                    cb_uri_changed, c);
	CB(c->wv, "scroll-event",                   cb_wv_hid, c);
	CB(c->wv, "web-process-crashed",            cb_wv_crashed, c);

	/* get webkit settings */
	c->settings = webkit_web_view_get_settings(WEBKIT_WEB_VIEW(c->wv));

	/* load config into webkit settings */
	if (CFG_S(UserAgent) != NULL)
		g_object_set(
		    G_OBJECT(c->settings), "user-agent", CFG_S(UserAgent), NULL
		);
	if (CFG_L(AcceptedLanguages) != NULL)
		webkit_web_context_set_preferred_languages(
		    webkit_web_view_get_context(WEBKIT_WEB_VIEW(c->wv)),
		    CFG_L(AcceptedLanguages)
		);
	if (CFG_S(CookieFile) != NULL)
		webkit_cookie_manager_set_persistent_storage(
		    webkit_web_context_get_cookie_manager(
		        webkit_web_view_get_context(WEBKIT_WEB_VIEW(c->wv))
		    ),
		    CFG_S(CookieFile),
		    WEBKIT_COOKIE_PERSISTENT_STORAGE_TEXT
		);
	if (CFG_S(ProxyUri) != NULL)
		webkit_website_data_manager_set_network_proxy_settings(
		    webkit_web_view_get_website_data_manager(WEBKIT_WEB_VIEW(c->wv)),
		    WEBKIT_NETWORK_PROXY_MODE_CUSTOM,
		    webkit_network_proxy_settings_new(
		        CFG_S(ProxyUri), CFG_L(ProxyIgnore)
		    )
		);
	webkit_web_view_set_zoom_level(
	    WEBKIT_WEB_VIEW(c->wv), CFG_F(ZoomLevel)
	);
	webkit_settings_set_auto_load_images(
	    c->settings, !CFG_B(DisableAutoLoadImages)
	);
	webkit_settings_set_default_font_family(
	    c->settings, CFG_S(DefaultFont)
	);
	webkit_settings_set_default_font_size(
	    c->settings, CFG_I(DefaultFontSize)
	);
	webkit_settings_set_enable_developer_extras(
	    c->settings, CFG_B(DeveloperExtras)
	);
	webkit_settings_set_enable_smooth_scrolling(
	    c->settings, CFG_B(SmoothScrolling)
	);
	webkit_settings_set_enable_write_console_messages_to_stdout(
	    c->settings, CFG_B(ConsoleToStdout)
	);
	webkit_settings_set_monospace_font_family(
	    c->settings, CFG_S(MonospaceFont)
	);
	webkit_settings_set_sans_serif_font_family(
	    c->settings, CFG_S(SansSerifFont)
	);
	webkit_settings_set_serif_font_family(
	    c->settings, CFG_S(SerifFont)
	);

	/* create entry */
	c->entry = gtk_entry_new();
	CB(c->entry, "key-press-event",             cb_entry_hid, c);
	CB(c->entry, "icon-release",                cb_entry_icon_hid, c);
	/* entry must exist before first call */
	set_javascript_policy(c, JSP_ENABLE);

	/* create vertical box to store the web view and the entry */
	c->vbx = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
	gtk_box_pack_start(GTK_BOX(c->vbx), c->wv, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(c->vbx), c->entry, FALSE, FALSE, 0);
	gtk_container_set_focus_child(GTK_CONTAINER(c->vbx), c->wv);

	/* create tab containing an icon and a label */
	c->tab_icon = gtk_image_new_from_icon_name(ICON_GLOBE, GTK_ICON_SIZE_MENU);
	c->tab_label = gtk_label_new(__NAME__);
	gtk_label_set_ellipsize(GTK_LABEL(c->tab_label), PANGO_ELLIPSIZE_END);
	gtk_widget_set_halign(GTK_WIDGET(c->tab_label), GTK_ALIGN_START);
	gtk_widget_set_hexpand(GTK_WIDGET(c->tab_label), TRUE);
	gtk_widget_set_has_tooltip(GTK_WIDGET(c->tab_label), TRUE);
	/* pack icon and label in a horizontal box */
	t = gtk_box_new(
	    GTK_ORIENTATION_HORIZONTAL, 5 * gtk_widget_get_scale_factor(mw.win)
	);
	gtk_box_pack_start(GTK_BOX(t), c->tab_icon, FALSE, FALSE, 0);
	gtk_box_pack_start(
	    GTK_BOX(t),
	    c->tab_label,
	    TRUE,
	    TRUE,
	    3 * gtk_widget_get_scale_factor(mw.win)
	);
	/* back the tab inside an event box to enable scroll and button events */
	e = gtk_event_box_new();
	gtk_container_add(GTK_CONTAINER(e), t);
	gtk_widget_add_events(e, GDK_SCROLL_MASK);
	CB(e, "button-release-event",               cb_tab_hid, c);
	CB(e, "scroll-event",                       cb_tab_hid, c);
	/* store a (non client) reference to the label for later use */
	g_object_set_data(G_OBJECT(e), __NAME__"-tab-label", c->tab_label);
	/* show tab */
	gtk_widget_show_all(e);

	/* append everything reorderable to the notebook after current page */
	gtk_notebook_insert_page(
	    GTK_NOTEBOOK(mw.nb),
	    c->vbx,
	    e,
	    gtk_notebook_get_current_page(GTK_NOTEBOOK(mw.nb)) + 1
	);
	gtk_notebook_set_tab_reorderable(GTK_NOTEBOOK(mw.nb), c->vbx, TRUE);

	/* manage which tab should be focused */
	c->focus_new_tab = focus_tab;
	if (show)
		show_web_view(c);
	else
		CB(c->wv, "ready-to-show",              cb_wv_show, c);

	/* finally load the uri */
	if (uri != NULL) {
		webkit_web_view_load_uri(
		    WEBKIT_WEB_VIEW(c->wv), (u = resolve_uri(uri))
		);
		g_free(u);
	}

	gl.clients++;
	return c;
}

void
client_destroy(struct Client *c)
{
	int                     i;

	/* disconnect all handlers */
	g_signal_handlers_disconnect_by_data(G_OBJECT(c->wv), c);
	g_signal_handlers_disconnect_by_data(G_OBJECT(c->wv), NULL);

	if ((i = gtk_notebook_page_num(GTK_NOTEBOOK(mw.nb), c->vbx)) != -1)
		gtk_notebook_remove_page(GTK_NOTEBOOK(mw.nb), i);
	else
		fprintf(stderr, __NAME__": Page does not exist in notebook\n");

	free(c);

	if (--gl.clients == 0)
		quit();
}

gboolean
command(struct Client *c, const gchar* t)
{
	if (t[0] == '/') {          /* in-page search */
		g_free(gl.search_text);
		gl.search_text = g_strdup(t + 1);
		search(c, IPS_INIT);
		return TRUE;
	} else if (t[0] == 'q') {   /* quit (vim-like) */
		quit();
		return TRUE;
	}

	return FALSE;
}

void
create_context_menu(struct Client *c, WebKitContextMenu *context_menu,
                    WebKitHitTestResult *hit_test_result)
{
	GAction                *a;                  /* action: freed when quit */
	if (webkit_hit_test_result_context_is_link(hit_test_result)) {
		CB(
		    (a = (GAction*)g_simple_action_new("open-external", NULL)),
		    "activate",
		    cb_open_external,
		    c
		);
		webkit_context_menu_insert(
		    context_menu,
		    webkit_context_menu_item_new_from_gaction(
		        a, "Open link externally", NULL
		    ),
		    2
		);
	}
}

void
die(const char *fmt, ...)
{
	va_list                 ap;

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	exit(EXIT_FAILURE);
}

gboolean
ipc_request(GIOChannel *src, GIOCondition condition, gpointer data)
{
	gchar                  *u = NULL;           /* uri received */

	(void)condition;        /* not used */
	(void)data;             /* not used */

	g_io_channel_read_line(src, &u, NULL, NULL, NULL);
	if (u) {
		g_strstrip(u);
		client_create(u, NULL, TRUE, TRUE);
		g_free(u);
	}
	return TRUE;
}

ssize_t
ipc_send(char *data)
{
	size_t                  s = 0,              /* data amount sent */
	                        l = strlen(data);   /* data length */
	ssize_t                 w;                  /* data written */

	while (s < l) { /* write until everything is sent */
		if ((w = write(gl.ipc_pipe_fd, data + s, l - s)) == -1) {
			if (errno == EINTR)
				continue;       /* EINTR: so retry */
			perror(__NAME__": Error writing to fifo");
			return w;
		} else if (w == 0) {
			return 0;
		}
		s += w;
	}

	return s;
}

void
ipc_setup(void)
{
	gchar                  *f,                  /* fifo file name */
	                       *p;                  /* fifo file path */

	/* only setup for ipc clients */
	if (gl.ipc != IPC_CLIENT)
		return;

	/* create fifo file path and allow for several ipc setups by naming them */
	f = g_strdup_printf("%s_%s.fifo", __NAME__, CFG_S(FifoName));
	p = g_build_filename(g_get_user_runtime_dir(), f, NULL);

	/* fail hard if fifo cannot be created */
	if (!g_file_test(p, G_FILE_TEST_EXISTS) && mkfifo(p, 0600) == -1)
		die("%s: fatal: mkfifo failed\n", __NAME__);

	/* if no one is listening create the ipc host instance */
	if ((gl.ipc_pipe_fd = open(p, O_WRONLY | O_NONBLOCK)) == -1) {
		g_io_add_watch(
		    g_io_channel_new_file(p, "r+", NULL),
		    G_IO_IN,
		    (GIOFunc)ipc_request,
		    NULL
		);
		gl.ipc = IPC_HOST;  /* convert client to host */
	}

	g_free(f);
	g_free(p);
}

gboolean
key_common(struct Client *c, GdkEventKey *event)
{
	int                     i,
	                        m;
	gchar                  *u;

	/* only handle key presses using the alt key */
	if (!(event->state & GDK_MOD1_MASK))
		return FALSE;

	/* key presses using ctrl+alt keys */
	if (event->state & GDK_CONTROL_MASK) {
		switch (event->keyval) {
		case GDK_KEY_J:
		case GDK_KEY_K: /* move page/tab backwards or forwards in stack */
			i = gtk_notebook_page_num(GTK_NOTEBOOK(mw.nb), c->vbx);
			m = gtk_notebook_get_n_pages(GTK_NOTEBOOK(mw.nb)) - 1;
			i += event->keyval == GDK_KEY_J
			    ? i == 0 ? 0 : -1
			    : i == m ? m : 1;
			gtk_notebook_reorder_child(GTK_NOTEBOOK(mw.nb), c->vbx, i);
			return TRUE;
		case GDK_KEY_j: /* go to previous page/tab */
			gtk_notebook_prev_page(GTK_NOTEBOOK(mw.nb));
			return TRUE;
		case GDK_KEY_k: /* go to next page/tab */
			gtk_notebook_next_page(GTK_NOTEBOOK(mw.nb));
			return TRUE;
		}
	}

	switch (event->keyval) {
	case GDK_KEY_q: /* destroy client and close tab */
		client_destroy(c);
		return TRUE;
	case GDK_KEY_w: /* go to home page */
		webkit_web_view_load_uri(
		    WEBKIT_WEB_VIEW(c->wv), (u = resolve_uri(CFG_S(HomeUri)))
		);
		g_free(u);
		return TRUE;
	case GDK_KEY_t: /* create new client/tab */
		c = client_create((u = resolve_uri(CFG_S(HomeUri))), NULL, TRUE, TRUE);
		g_free(u);
		gtk_widget_grab_focus(c->entry);
		return TRUE;
	case GDK_KEY_r: /* reload, while bypassing cache */
		webkit_web_view_reload_bypass_cache(WEBKIT_WEB_VIEW(c->wv));
		return TRUE;
	case GDK_KEY_o: /* focus entry */
		gtk_widget_grab_focus(c->entry);
		return TRUE;
	case GDK_KEY_n: /* search forward */
		search(c, IPS_FORWARD);
		return TRUE;
	case GDK_KEY_N: /* search backward */
		search(c, IPS_BACKWARD);
		return TRUE;
	case GDK_KEY_slash: /* initiate in-page search */
		gtk_widget_grab_focus(c->entry);
		gtk_entry_set_text(GTK_ENTRY(c->entry), ":/");
		gtk_editable_set_position(GTK_EDITABLE(c->entry), -1);
		return TRUE;
	case GDK_KEY_b: /* toggle tab visibility */
		gtk_notebook_set_show_tabs(
		    GTK_NOTEBOOK(mw.nb),
		    !gtk_notebook_get_show_tabs(GTK_NOTEBOOK(mw.nb))
		);
		return TRUE;
	case GDK_KEY_1:
	case GDK_KEY_2:
	case GDK_KEY_3:
	case GDK_KEY_4:
	case GDK_KEY_5:
	case GDK_KEY_6:
	case GDK_KEY_7:
	case GDK_KEY_8:
	case GDK_KEY_9: /* go to nth tab */
		gtk_notebook_set_current_page(GTK_NOTEBOOK(mw.nb), event->keyval-0x31);
		return TRUE;
	case GDK_KEY_H: /* go back in history */
		webkit_web_view_go_back(WEBKIT_WEB_VIEW(c->wv));
		return TRUE;
	case GDK_KEY_L: /* go forward in history */
		webkit_web_view_go_forward(WEBKIT_WEB_VIEW(c->wv));
		return TRUE;
	case GDK_KEY_s: /* toggle javascript (on or off) */
		set_javascript_policy(c, JSP_TOGGLE);
		return TRUE;
	case GDK_KEY_c: /* toggle tls error policy (ignore or fail) */
		toggle_tls_error_policy(c);
		return TRUE;
	case GDK_KEY_h:
	case GDK_KEY_j:
	case GDK_KEY_k:
	case GDK_KEY_l: /* in-page movement by emulating arrow keys... */
	case GDK_KEY_g:
	case GDK_KEY_G: /* ... as well as home and end */
		event->keyval /* translate key press accordingly */
		    = event->keyval == GDK_KEY_h ? GDK_KEY_Left
		    : event->keyval == GDK_KEY_j ? GDK_KEY_Down
		    : event->keyval == GDK_KEY_k ? GDK_KEY_Up
		    : event->keyval == GDK_KEY_l ? GDK_KEY_Right
		    : event->keyval == GDK_KEY_g ? GDK_KEY_Home
		    : /*                        */ GDK_KEY_End;
		gtk_propagate_event(GTK_WIDGET(c->wv), (GdkEvent*)event);
		return TRUE;
	case GDK_KEY_p: /* open print dialog */
		webkit_print_operation_run_dialog(
		    webkit_print_operation_new(WEBKIT_WEB_VIEW(c->wv)),
		    GTK_WINDOW(mw.win)
		);
		return TRUE;
	}
	return FALSE;
}

gboolean
key_entry(struct Client *c, GdkEventKey *event)
{
	const gchar            *t,
	                       *l;
	gchar                  *u;
	int                     p,
	                        m;

	/* handle key presses using the alt key (takes precedence over common) */
	if (event->state & GDK_MOD1_MASK) {
		switch (event->keyval) { /* movements inside entry */
		case GDK_KEY_l:
		case GDK_KEY_h:
			p = gtk_editable_get_position(GTK_EDITABLE(c->entry));
			m = gtk_entry_get_text_length(GTK_ENTRY(c->entry));
			p += event->keyval == GDK_KEY_l
				? m > p ? 1 : 0     /* move right */
				: 0 < p ? -1 : 0;   /* move left */
			gtk_editable_set_position(GTK_EDITABLE(c->entry), p);
			return TRUE;
		}
	}

	/* handle common key presses */
	if (key_common(c, event))
		return TRUE;

	/* handle any key press (not just using the alt key) */
	switch (event->keyval) {
	case GDK_KEY_KP_Enter:
	case GDK_KEY_Return:
		gtk_widget_grab_focus(c->wv);
		if ((t = gtk_entry_get_text(GTK_ENTRY(c->entry))) == NULL)
			return TRUE;
		if (t[0] != ':') { /* if not a command */
			/* store current uri before loading new uri */
			l = WV_GET_URI(c->wv);
			webkit_web_view_load_uri(
			    WEBKIT_WEB_VIEW(c->wv), (u = resolve_uri(t))
			);
			g_free(u);
			/* fix: notify::uri won't be raised if current uri is unchanged */
			if (g_strcmp0(l, WV_GET_URI(c->wv)) == 0)
				uri_changed(c);
			return TRUE;
		} else if (command(c, t + 1)) { /* if command */
			return TRUE;
		} /* FALL THROUGH */
	case GDK_KEY_Escape:
		gtk_widget_grab_focus(c->wv);
		t = WV_GET_URI(c->wv);
		gtk_entry_set_text(GTK_ENTRY(c->entry), (t == NULL ? __NAME__ : t));
		gtk_editable_set_position(GTK_EDITABLE(c->entry), -1);
		return TRUE;
	}
	return FALSE;
}

gboolean
key_tab(struct Client *c, GdkEvent *event)
{
	GdkScrollDirection      d;

	if (event->type == GDK_BUTTON_RELEASE) {
		if (((GdkEventButton *)event)->button == 2) { /* scroll wheel click */
			client_destroy(c);
			return TRUE;
		}
	} else if (event->type == GDK_SCROLL) { /* scroll wheel selection */
		gdk_event_get_scroll_direction(event, &d);
		if (d == GDK_SCROLL_UP)
			gtk_notebook_next_page(GTK_NOTEBOOK(mw.nb));
		else if (d == GDK_SCROLL_DOWN)
			gtk_notebook_prev_page(GTK_NOTEBOOK(mw.nb));
		return TRUE;
	}
	return FALSE;
}

gboolean
key_web_view(struct Client *c, GdkEvent *event)
{
	gdouble                 dx,                 /* scroll x-delta */
	                        dy;                 /* scroll y-delta */

	/* handle common key presses */
	if (event->type == GDK_KEY_PRESS && key_common(c, (GdkEventKey *)event))
		return TRUE;

	/* escape key: stop web rendering */
	if (event->type == GDK_KEY_PRESS
	    && ((GdkEventKey *)event)->keyval == GDK_KEY_Escape) {
		webkit_web_view_stop_loading(WEBKIT_WEB_VIEW(c->wv));
		gtk_entry_set_progress_fraction(GTK_ENTRY(c->entry), 0);
	/* mouse button events */
	} else if (event->type == GDK_BUTTON_RELEASE) {
		switch (((GdkEventButton *)event)->button) {
		case 2:  /* mouse middle button: open in new tab */
			if (c->hover_uri != NULL) {
				client_create(c->hover_uri, NULL, TRUE, FALSE);
				return TRUE;
			}
			break;
		case 8:  /* mouse button back: go back in history */
			webkit_web_view_go_back(WEBKIT_WEB_VIEW(c->wv));
			return TRUE;
		case 9:  /* mouse button forward: go forward in history */
			webkit_web_view_go_forward(WEBKIT_WEB_VIEW(c->wv));
			return TRUE;
		}
	/* scroll using alt pressed: zoom in or out */
	} else if (event->type == GDK_SCROLL
	           && (((GdkEventScroll *)event)->state & GDK_MOD1_MASK)) {
		gdk_event_get_scroll_deltas(event, &dx, &dy);
		webkit_web_view_set_zoom_level(
		    WEBKIT_WEB_VIEW(c->wv),
		    dx != 0
		    ? CFG_F(ZoomLevel)
		    : webkit_web_view_get_zoom_level(WEBKIT_WEB_VIEW(c->wv)) + -dy*0.1);
		return TRUE;
	}

	return FALSE;
}

void
load_changed(struct Client *c, WebKitLoadEvent event_type)
{
	switch (event_type) {
	/* when page load starts, reset everything */
	case WEBKIT_LOAD_STARTED:
		c->https = FALSE;
		if (c->error_page)
			c->error_page = FALSE;
		else
			g_clear_object(&c->failed_crt);
		break;
	/* when page load is redirected, continue as usual */
	case WEBKIT_LOAD_REDIRECTED:
		break;
	/* when page load is committed, get https and tls state */
	case WEBKIT_LOAD_COMMITTED:
		c->https = webkit_web_view_get_tls_info(
		    WEBKIT_WEB_VIEW(c->wv), &c->crt, &c->tls_error
		);
		break;
	/* when page load is finished, run all user scripts */
	case WEBKIT_LOAD_FINISHED:
		load_user_styles(WEBKIT_WEB_VIEW(c->wv));
		run_user_scripts(WEBKIT_WEB_VIEW(c->wv));
		break;
	}

	update_title(c);
}

void
load_configuration(void)
{
	const gchar            *e;                  /* environment variable name */
	int                     i;

	/* set global defaults */
	gl.clients              = 0;                /* client counter */
	gl.ipc                  = IPC_CLIENT;       /* ipc type */
	gl.ipc_pipe_fd          = 0;                /* ipc pipe file descriptor */
	gl.search_text          = NULL;             /* in page search phrase */
	gl.state_lock           = TRUE;             /* prevents state save */

	/* load default configuration */
	for (i = 0; i < LastConfig; i++) {
		if (cfg[i].e && (e = g_getenv(cfg[i].e)) != NULL) {
			switch (cfg[i].t) {
			case CFG_INT:       cfg[i].v.i = atoi(e); break;
			case CFG_FLOAT:     cfg[i].v.f = atof(e); break;
			case CFG_BOOL:      cfg[i].v.b = TRUE; break;
			case CFG_STRING:    cfg[i].v.s = g_strdup(e); break;
			case CFG_LIST:      cfg[i].v.l = g_strsplit(e, ",", -1); break;
			}
		}
	}
}

void
load_state(void)
{
	char                   *u = NULL;           /* line/uri holder */
	size_t                  l = URI_MAX;        /* length */
	FILE                   *fp = NULL;          /* file pointer */

	/* only load states for ipc host */
	if (CFG_S(StateFile) == NULL || gl.ipc != IPC_HOST)
		return;

	if ((fp = fopen(CFG_S(StateFile), "r")) == NULL) {
		perror(__NAME__": Error opening state file");
		return;
	}

	if (!(u = (char *)malloc(l * sizeof(char))))
		die("%s: fatal: malloc failed\n", __NAME__);

	/* load uris into new clients */
	while (getline(&u, &l, fp) >= 0)
		client_create(u, NULL, TRUE, TRUE);

	/* select first tab if any */
	if (gl.clients > 0)
		gtk_notebook_set_current_page(GTK_NOTEBOOK(mw.nb), 0);

	/* enable state save */
	gl.state_lock = FALSE;

	fclose(fp);
	free(u);
}

void
load_user_styles(WebKitWebView *web_view)
{
	gchar                  *c = NULL,           /* file content */
	                       *p = NULL,           /* path (file) */
	                       *b = NULL;           /* base directory (scripts) */
	const gchar            *e = NULL;           /* directory file entry */
	GDir                   *s = NULL;           /* user style directory */

	b = g_build_filename(g_get_user_config_dir(), __NAME__, CFG_S(UcDir), NULL);
	if ((s = g_dir_open(b, 0, NULL)) == NULL) {
		g_free(b);
		return;
	}

	webkit_user_content_manager_remove_all_style_sheets(
	    webkit_web_view_get_user_content_manager(web_view)
	);

	while ((e = g_dir_read_name(s)) != NULL) {
		if (g_str_has_suffix((p = g_build_filename(b, e, NULL)), ".css")
		    && g_file_get_contents(p, &c, NULL, NULL)) {
			webkit_user_content_manager_add_style_sheet(
			    webkit_web_view_get_user_content_manager(web_view),
			    webkit_user_style_sheet_new(
			        c,
			        WEBKIT_USER_CONTENT_INJECT_ALL_FRAMES,
			        WEBKIT_USER_STYLE_LEVEL_USER,
			        NULL,
			        NULL
			    )
			);
			g_free(c);
		}
		g_free(p);
	}

	g_dir_close(s);
	g_free(b);
}

void
main_window_setup(void)
{
	GtkWidget              *v;                  /* vertical box */

	/* define window elements */
	mw.win                  = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	mw.nb                   = gtk_notebook_new();
	mw.dbx                  = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
	v                       = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);

	/* window settings */
	gtk_window_set_default_size(GTK_WINDOW(mw.win), 800, 600);
	gtk_window_set_title(GTK_WINDOW(mw.win), __NAME__);

	/* notebook settings */
	gtk_notebook_set_scrollable(GTK_NOTEBOOK(mw.nb), TRUE);
	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(mw.nb), GTK_POS_TOP);

	/* pack everything in a vertical box */
	gtk_box_pack_start(GTK_BOX(v), mw.nb, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(v), mw.dbx, FALSE, FALSE, 0);
	gtk_container_add(GTK_CONTAINER(mw.win), v);

	/* connect signal callbacks */
	CB(mw.win,              "destroy",          cb_quit, NULL);
	CB(mw.nb,               "switch-page",      cb_notebook_switch_page, NULL);
	CB(mw.nb,               "page-added",       cb_notebook_modified, NULL);
	CB(mw.nb,               "page-removed",     cb_notebook_modified, NULL);
	CB(mw.nb,               "page-reordered",   cb_notebook_modified, NULL);
}

void
open_external(struct Client *c)
{
	char                    b[URI_MAX];         /* command line buffer */
	gchar                  *p = NULL;           /* external handler path */
	p = g_build_filename(
	    g_get_user_config_dir(), __NAME__, CFG_S(ExternalHandlerFile), NULL
	);
	sprintf(b, "%s %s &", p, gtk_entry_get_text(GTK_ENTRY(c->entry)));
	g_free(p);

	system(b);
}

void
prepare_download(WebKitDownload *d, gchar *suggested_filename)
{
	gchar                  *s = g_strdup(suggested_filename),
	                       *p = NULL,           /* proposed path */
	                       *f = NULL,           /* final path */
	                       *u = NULL;           /* uri */
	GtkWidget              *b;                  /* button */
	int                     i;

	/* make suggested filename safe, by replacing directory separators */
	for (i = 0; s[i]; i++)
		if (s[i] == G_DIR_SEPARATOR)
			s[i] = '_';

	p = g_build_filename(CFG_S(DownloadDirectory), s, NULL);
	f = g_strdup(p);

	/* check for a free suffix */
	for (i = 1; g_file_test(f, G_FILE_TEST_EXISTS) && i < SUFFIX_MAX; i++) {
		g_free(f);
		f = g_strdup_printf("%s.%d", p, i);
	}

	if (i < SUFFIX_MAX) {
		u = g_filename_to_uri(f, NULL, NULL);
		webkit_download_set_destination(d, u);

		b = gtk_button_new_with_label(s);
		gtk_button_set_always_show_image(GTK_BUTTON(b), TRUE);
		gtk_button_set_image(
		    GTK_BUTTON(b),
		    gtk_image_new_from_icon_name(ICON_DOWNLOAD, GTK_ICON_SIZE_BUTTON)
		);
		gtk_widget_set_margin_end(b, 1);
		gtk_box_pack_start(GTK_BOX(mw.dbx), b, FALSE, FALSE, 0);
		gtk_box_reorder_child(GTK_BOX(mw.dbx), b, 0);
		gtk_widget_show_all(mw.dbx);

		CB(d, "notify::estimated-progress", cb_download_changed_progress, b);
		CB(d, "finished", cb_download_finished, b);
		g_object_ref(d);
		CB(b, "clicked", cb_download_cancel, d);
	} else { /* could not find a free suffix under the SUFFIX_MAX limit */
		fprintf(stderr, __NAME__": Limit reached for filename suffix\n");
		webkit_download_cancel(d);
	}

	g_free(p);
	g_free(s);
	g_free(f);
	g_free(u);
}

void
quit(void)
{
	save_state();
	gtk_main_quit();
}

void
render_tls_error(struct Client *c, gchar *uri, GTlsCertificate *crt,
                 GTlsCertificateFlags cert_flags)
{
	GString                *m;                  /* message (error) */
	gchar                  *h,                  /* html code */
	                       *p;                  /* pem block */

	m                       = g_string_new(NULL);
	c->failed_crt           = g_object_ref(crt);
	c->tls_error            = cert_flags;
	c->error_page           = TRUE;

	/* translate all flags to messages */
	if (c->tls_error & G_TLS_CERTIFICATE_UNKNOWN_CA)
		g_string_append(m, MSG_TLS_CERTIFICATE_UNKNOWN_CA);
	if (c->tls_error & G_TLS_CERTIFICATE_BAD_IDENTITY)
		g_string_append(m, MSG_TLS_CERTIFICATE_BAD_IDENTITY);
	if (c->tls_error & G_TLS_CERTIFICATE_NOT_ACTIVATED)
		g_string_append(m, MSG_TLS_CERTIFICATE_NOT_ACTIVATED);
	if (c->tls_error & G_TLS_CERTIFICATE_EXPIRED)
		g_string_append(m, MSG_TLS_CERTIFICATE_EXPIRED);
	if (c->tls_error & G_TLS_CERTIFICATE_REVOKED)
		g_string_append(m, MSG_TLS_CERTIFICATE_REVOKED);
	if (c->tls_error & G_TLS_CERTIFICATE_INSECURE)
		g_string_append(m, MSG_TLS_CERTIFICATE_INSECURE);
	if (c->tls_error & G_TLS_CERTIFICATE_GENERIC_ERROR)
		g_string_append(m, MSG_TLS_CERTIFICATE_GENERIC_ERROR);

	/* construct html code and load it */
	g_object_get(crt, "certificate-pem", &p, NULL);
	h = g_strdup_printf(TLS_MSG_FORMAT, uri, m->str, p);
	webkit_web_view_load_alternate_html(WEBKIT_WEB_VIEW(c->wv), h, uri, NULL);

	g_string_free(m, TRUE);
	g_free(h);
	g_free(p);
}

void
run_user_scripts(WebKitWebView *web_view)
{
	gchar                  *c = NULL,           /* file content */
	                       *p = NULL,           /* path (file) */
	                       *b = NULL;           /* base directory (scripts) */
	const gchar            *e = NULL;           /* directory file entry */
	GDir                   *s = NULL;           /* user script directory */

	b = g_build_filename(g_get_user_config_dir(), __NAME__, CFG_S(UsDir), NULL);
	if ((s = g_dir_open(b, 0, NULL)) == NULL) {
		g_free(b);
		return;
	}

	while ((e = g_dir_read_name(s)) != NULL) {
		if (g_str_has_suffix((p = g_build_filename(b, e, NULL)), ".js")
		    && g_file_get_contents(p, &c, NULL, NULL)) {
			webkit_web_view_run_javascript(web_view, c, NULL, NULL, NULL);
			g_free(c);
		}
		g_free(p);
	}

	g_dir_close(s);
	g_free(b);
}

void
save_history(const gchar *t)
{
	FILE                   *fp;

	if (CFG_S(HistoryFile) == NULL)
		return;

	if ((fp = fopen(CFG_S(HistoryFile), "a")) != NULL) {
		fprintf(fp, "%s\n", t);
		fclose(fp);
	} else {
		perror(__NAME__": Error opening history file");
	}
}

void
save_state(void)
{
	GList                  *c;                  /* child (notebook page) */
	int                     i,
	                        l;                  /* length (number of tabs) */
	FILE                   *fp;

	/* only save state for ipc host, when state isn't locked for loading */
	if (CFG_S(StateFile) == NULL || gl.ipc != IPC_HOST || gl.state_lock)
		return;

	if ((fp = fopen(CFG_S(StateFile), "w+")) == NULL) {
		perror(__NAME__": Error opening state file");
		return;
	}

	/* save uri from each tab into state file */
	for (i = 0, l = gtk_notebook_get_n_pages(GTK_NOTEBOOK(mw.nb)); i < l; i++) {
		c = gtk_container_get_children(
		    GTK_CONTAINER(gtk_notebook_get_nth_page(GTK_NOTEBOOK(mw.nb), i))
		);
		if (strcmp(gtk_widget_get_name((c->data)), "WebKitWebView") == 0)
			fprintf(fp, "%s\n", WV_GET_URI(c->data));
	}

	fclose(fp);
}

void
search(struct Client *c, enum inpage_search_type type)
{
	if (gl.search_text == NULL)
		return;

	switch (type) {
	case IPS_INIT:
		webkit_find_controller_search(
		    webkit_web_view_get_find_controller(WEBKIT_WEB_VIEW(c->wv)),
		    gl.search_text,
		    WEBKIT_FIND_OPTIONS_CASE_INSENSITIVE
		    | WEBKIT_FIND_OPTIONS_WRAP_AROUND,
		    G_MAXUINT
		);
		break;
	case IPS_FORWARD:
		webkit_find_controller_search_next(
		    webkit_web_view_get_find_controller(WEBKIT_WEB_VIEW(c->wv))
		);
		break;
	case IPS_BACKWARD:
		webkit_find_controller_search_previous(
		    webkit_web_view_get_find_controller(WEBKIT_WEB_VIEW(c->wv))
		);
		break;
	}
}

void
set_default_web_context(void)
{
	gchar                  *p;                  /* web extensions path */
	WebKitWebContext       *c;                  /* web context */

	/* no context needed for clients */
	if (gl.ipc == IPC_CLIENT)
		return;

	c = webkit_web_context_get_default();
	webkit_web_context_set_process_model(
	    c, WEBKIT_PROCESS_MODEL_MULTIPLE_SECONDARY_PROCESSES
	);

	p = g_build_filename(g_get_user_config_dir(), __NAME__, CFG_S(WeDir), NULL);
#if GTK_CHECK_VERSION(3, 98, 0) /* seems to be fixed in 3.98.0, we'll see */
	webkit_web_context_set_sandbox_enabled(c, TRUE);
	webkit_web_context_add_path_to_sandbox(c, p, TRUE);
#else
	webkit_web_context_set_web_extensions_directory(c, p);
#endif

	CB(c, "download-started", cb_download_start, NULL);
	webkit_web_context_set_favicon_database_directory(c, NULL);

	g_free(p);
}

void
set_hover_uri(struct Client *c, WebKitHitTestResult *hit_test_result)
{
	const char             *t;                  /* entry text holder */

	g_free(c->hover_uri);

	/* only display hovered links */
	if (webkit_hit_test_result_context_is_link(hit_test_result)) {
		t = webkit_hit_test_result_get_link_uri(hit_test_result);
		c->hover_uri = g_strdup(t);
	} else {
		t = WV_GET_URI(c->wv);
		c->hover_uri = NULL;
	}

	if (!gtk_widget_is_focus(c->entry))
		gtk_entry_set_text(GTK_ENTRY(c->entry), t);
}

void
set_javascript_policy(struct Client *c, enum javascript_policy policy)
{
	webkit_settings_set_enable_javascript(c->settings, policy == JSP_TOGGLE
	    ? !(webkit_settings_get_enable_javascript(c->settings))
	    : policy
	);
	webkit_web_view_reload_bypass_cache(WEBKIT_WEB_VIEW(c->wv));
	if ((webkit_settings_get_enable_javascript(c->settings)))
		gtk_entry_set_icon_from_icon_name(
		    GTK_ENTRY(c->entry), GTK_ENTRY_ICON_SECONDARY, ICON_JS_ON
		);
	else
		gtk_entry_set_icon_from_icon_name(
		    GTK_ENTRY(c->entry), GTK_ENTRY_ICON_SECONDARY, ICON_JS_OFF
		);
}

void
set_window_title(gint i)
{
	GtkWidget              *c;                  /* notebook child/page */

	if ((c = gtk_notebook_get_nth_page(GTK_NOTEBOOK(mw.nb), i)) == NULL)
		return;

	gtk_window_set_title(GTK_WINDOW(mw.win),
	    gtk_label_get_text(
	        GTK_LABEL((GtkWidget *)g_object_get_data(
	            G_OBJECT(gtk_notebook_get_tab_label(GTK_NOTEBOOK(mw.nb), c)),
	            __NAME__"-tab-label"
	        ))
	    )
	);
}

void
show_web_view(struct Client *c)
{
	gint                    i;

	gtk_widget_show_all(mw.win);

	if (c->focus_new_tab) {
		if ((i = gtk_notebook_page_num(GTK_NOTEBOOK(mw.nb), c->vbx)) != -1)
			gtk_notebook_set_current_page(GTK_NOTEBOOK(mw.nb), i);
		gtk_widget_grab_focus(c->wv);
	}
}

void
toggle_tls_error_policy(struct Client *c)
{
	webkit_website_data_manager_set_tls_errors_policy(
	    webkit_web_view_get_website_data_manager(WEBKIT_WEB_VIEW(c->wv)),
	    !webkit_website_data_manager_get_tls_errors_policy(
	        webkit_web_view_get_website_data_manager(WEBKIT_WEB_VIEW(c->wv))
	    )
	);
	webkit_web_view_reload(WEBKIT_WEB_VIEW(c->wv));
}

gchar *
resolve_uri(const gchar *t)
{
	gchar                  *u,                  /* uri to return */
	                       *l = NULL;           /* temporary string */
	const gchar            *s;

	/* check if valid scheme, and if so just create a copy of the text */
	if ((s = g_uri_peek_scheme(t)) && g_strv_contains(CFG_L(UriSchemes), s))
		u = g_strdup(t);
	/* if path is local, use the file scheme */
	else if ((l = realpath(t, NULL)) != NULL)
		u = g_strdup_printf("file://%s", l);
	/* else, check if the text can be interpreted as a valid https uri; it's
	 * not enough to check if uri is valid - check for period and no spaces */
	else if (strchr(t, '.') && !strchr(t, ' ')
	         && g_uri_is_valid((l = g_strdup_printf("https://%s", t)), 0, NULL))
		u = g_strdup(l);
	/* fallback to web search, using a specified search engine */
	else
		u = g_strdup_printf(CFG_S(SearchEngineUriFormat), t);

	g_free(l);
	return u; /* return a pointer that the caller is responsible for freeing */
}

void
update_download_button(WebKitDownload *d, GtkButton *btn, gboolean done)
{
	gdouble                 p,                  /* percent completed */
	                        l;                  /* responce length (bytes) */
	gchar                  *t,                  /* text holder*/
	                       *f,                  /* file name */
	                       *b;                  /* base name */

	if ((f = g_filename_from_uri(webkit_download_get_destination(d), NULL, NULL)
	    ) == NULL)
		return; /* should never happen, but just in case... */

	p = webkit_download_get_estimated_progress(d);
	p = (p > 1 ? 1 : p < 0 ? 0 : p) * 100;
	l = webkit_uri_response_get_content_length(webkit_download_get_response(d));
	b = g_path_get_basename(f);

	gtk_button_set_label(btn, !done /* set label based on progress */
		? (t = g_strdup_printf(" %s (%.0f%% of %.1f MB)", b, p, l / 1e6))
		: (t = g_strdup_printf(" %s", b))
	);

	g_free(t);
	g_free(f);
	g_free(b);
}

void
update_favicon(struct Client *c)
{
	cairo_surface_t        *f;  /* favicon */
	GdkPixbuf              *b,  /* pix buffer */
	                       *s;  /* pix buffer (scaled) */

	/* set fallback favicon */
	gtk_image_set_from_icon_name(
	    GTK_IMAGE(c->tab_icon), ICON_GLOBE, GTK_ICON_SIZE_SMALL_TOOLBAR
	);

	if ((f = webkit_web_view_get_favicon(WEBKIT_WEB_VIEW(c->wv))) == NULL)
		return;

	if ((b = gdk_pixbuf_get_from_surface(f, 0, 0,
	    cairo_image_surface_get_width(f),
	    cairo_image_surface_get_height(f))) == NULL)
		return;

	s = gdk_pixbuf_scale_simple(b,
	    16 * gtk_widget_get_scale_factor(c->tab_icon),
	    16 * gtk_widget_get_scale_factor(c->tab_icon),
	    GDK_INTERP_BILINEAR
	);
	gtk_image_set_from_pixbuf(GTK_IMAGE(c->tab_icon), s);

	g_object_unref(s);
	g_object_unref(b);
}

void
update_load_progress(struct Client *c)
{
	gdouble                 p;

	p = webkit_web_view_get_estimated_load_progress(WEBKIT_WEB_VIEW(c->wv));
	gtk_entry_set_progress_fraction(GTK_ENTRY(c->entry), (p == 1 ? 0 : p));
}

void
update_title(struct Client *c)
{
	const gchar            *t,
	                       *u;
	gchar                  *m;

	u = WV_GET_URI(c->wv);
	t = webkit_web_view_get_title(WEBKIT_WEB_VIEW(c->wv));

	/* title priority: title, url, __NAME__ */
	t = t == NULL || t[0] == 0 ? (u == NULL || u[0] == 0 ? __NAME__ : u) : t;

	/* set label markup and certificate icon based on tls status */
	if (c->failed_crt
	    || (c->https && c->tls_error != G_TLS_CERTIFICATE_NO_FLAGS)) {
		m = g_markup_printf_escaped(CFG_S(BadTlsTabFormat), t);
		gtk_entry_set_icon_from_icon_name(
		    GTK_ENTRY(c->entry), GTK_ENTRY_ICON_PRIMARY, ICON_BAD_TLS
		);
	} else {
		m = g_markup_printf_escaped(CFG_S(NormalTabFormat), t);
		if (c->https)
			gtk_entry_set_icon_from_icon_name(
			    GTK_ENTRY(c->entry), GTK_ENTRY_ICON_PRIMARY, ICON_TLS
			);
		else
			gtk_entry_set_icon_from_icon_name(
			    GTK_ENTRY(c->entry), GTK_ENTRY_ICON_PRIMARY, NULL
			);
	}

	gtk_label_set_markup(GTK_LABEL(c->tab_label), m);
	gtk_widget_set_tooltip_text(c->tab_label, t);
	set_window_title(gtk_notebook_get_current_page(GTK_NOTEBOOK(mw.nb)));
	g_free(m);
}

void
uri_changed(struct Client *c)
{
	const gchar            *t;

	/* make sure to not overwrite the "WEB PROCESS CRASHED" message */
	if ((t = WV_GET_URI(c->wv)) != NULL && strlen(t) > 0) {
		gtk_entry_set_text(GTK_ENTRY(c->entry), t);
		save_history(t);
		save_state();
	}
}

void
web_view_crashed(struct Client *c)
{
	gchar                  *t;

	gtk_entry_set_text(
	    GTK_ENTRY(c->entry),
	    (t = g_strdup_printf("WEB PROCESS CRASHED: %s", WV_GET_URI(c->wv)))
	);
	g_free(t);
}

int
main(int argc, char **argv)
{
	int                     opt,
	                        i;

	gtk_init(&argc, &argv);

	/* load default configuration before reading command-line arguments */
	load_configuration();

	while ((opt = getopt(argc, argv, "I")) != -1)
		switch (opt) {
		case 'I':
			gl.ipc = IPC_NONE;
			break;
		default:
			die("Usage: %s [OPTION]... [URI]...\n", __NAME__);
		}

	ipc_setup();
	set_default_web_context();
	main_window_setup();
	load_state();

	/* load a default home uri if no clients or arguments exist  */
	if (optind >= argc && gl.clients == 0)
		client_create(CFG_S(HomeUri), NULL, TRUE, TRUE);
	/* load remaining command line arguments as uris into new clients */
	else if (optind < argc)
		for (i = optind; i < argc; i++)
			client_create(argv[i], NULL, TRUE, TRUE);

	/* don't load gtk for ipc clients, they have done their bit */
	if (gl.ipc != IPC_CLIENT)
		gtk_main();

	exit(EXIT_SUCCESS);
}
