/**
 * Copyright (C) 2023 Chris Noxz
 * Author(s): Chris Noxz <chris@noxz.tech>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef BROWSER_H
#define BROWSER_H

#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <JavaScriptCore/JavaScript.h>
#include <gdk/gdkkeysyms.h>
#include <gio/gio.h>
#include <gtk/gtk.h>
#include <gtk/gtkx.h>
#include <webkit2/webkit2.h>

#define URI_MAX             2000
#define SUFFIX_MAX          100
#define ICON_DOWNLOAD       "network-receive-symbolic"
#define ICON_FINISHED       "emblem-default-symbolic"
#define ICON_GLOBE          "globe-symbolic"
#define ICON_TLS            "channel-secure-symbolic"
#define ICON_BAD_TLS        "channel-insecure-symbolic"
#define ICON_JS_ON          "emblem-ok-symbolic"
#define ICON_JS_OFF         "action-unavailable-symbolic"
#define CB(C, E, F, D)                                                         \
        {                                                                      \
            g_signal_connect(G_OBJECT(C), (E), G_CALLBACK(F), (D));            \
        }
#define CFG_B(X)            cfg[(X)].v.b
#define CFG_F(X)            cfg[(X)].v.f
#define CFG_I(X)            cfg[(X)].v.i
#define CFG_L(X)            (const gchar* const*)cfg[(X)].v.l
#define CFG_S(X)            cfg[(X)].v.s
#define WV_GET_URI(X)       webkit_web_view_get_uri(WEBKIT_WEB_VIEW((X)))

#define TLS_MSG_FORMAT      "<h2>Could not validate TLS for: %s</h2><pre>%s"   \
                            "</pre><pre>%s</pre>"
#define MSG_TLS_CERTIFICATE_UNKNOWN_CA                                         \
                            ">  The signing certificate authority is not known"\
                            ".<br>"
#define MSG_TLS_CERTIFICATE_BAD_IDENTITY                                       \
                            ">  The certificate does not match the expected"   \
                            "identity of the site that it was retrieved from." \
                            "<br>"
#define MSG_TLS_CERTIFICATE_NOT_ACTIVATED                                      \
                            ">  The certificate's activation time is still in" \
                            " the future.<br>"
#define MSG_TLS_CERTIFICATE_EXPIRED                                            \
                            ">  The certificate has expired.<br>"
#define MSG_TLS_CERTIFICATE_REVOKED                                            \
                            ">  The certificate has been revoked according to" \
                            " the GTlsConnection's certificate revocation list"\
                            ".<br>"
#define MSG_TLS_CERTIFICATE_INSECURE                                           \
                            ">  The certificate's algorithm is considered "    \
                            "insecure.<br>"
#define MSG_TLS_CERTIFICATE_GENERIC_ERROR                                      \
                            ">  Some error occurred validating the certificate"\
                            ".<br>"

enum javascript_policy {
	JSP_DISABLE             = 0,
	JSP_ENABLE              = 1,
	JSP_TOGGLE              = 2
};

enum inpage_search_type {
	IPS_INIT,
	IPS_FORWARD,
	IPS_BACKWARD
};

enum ipc_type {
	IPC_NONE,
	IPC_HOST,
	IPC_CLIENT
};

enum config_type {
	CFG_INT,
	CFG_FLOAT,
	CFG_BOOL,
	CFG_STRING,
	CFG_LIST,
};

enum config_name {
	AcceptedLanguages,
	BadTlsTabFormat,
	ConsoleToStdout,
	CookieFile,
	DefaultFont,
	DefaultFontSize,
	DeveloperExtras,
	DisableAutoLoadImages,
	DownloadDirectory,
	ExternalHandlerFile,
	FifoName,
	HistoryFile,
	HomeUri,
	MonospaceFont,
	NormalTabFormat,
	ProxyIgnore,
	ProxyUri,
	SansSerifFont,
	SearchEngineUriFormat,
	SerifFont,
	SmoothScrolling,
	StateFile,
	UriSchemes,
	UserAgent,
	UcDir,      /* user (cascading) style sheets directory */
	UsDir,      /* user scripts directory */
	WeDir,      /* web extension directory */
	ZoomLevel,
	/* must be last to represent number of config items */
	LastConfig
};

typedef union {
	gint                    i;              /* union integer */
	gdouble                 f;              /* union float/double */
	gboolean                b;              /* union boolean */
	gchar                  *s;              /* union string */
	gchar                 **l;              /* union string list/array */
} Arg;

struct Client
{
	GTlsCertificate        *crt;            /* last certificate */
	GtkWidget              *entry;          /* uri/command/entry */
	gboolean                error_page;     /* flag if error page */
	GTlsCertificate        *failed_crt;     /* last failed certificate */
	gboolean                focus_new_tab;  /* flag for focusing tab */
	gchar                  *hover_uri;      /* uri last hovered */
	gboolean                https;          /* flag if https site */
	WebKitSettings         *settings;       /* reference to webkit settings */
	GtkWidget              *tab_icon;       /* holder for favicon in tab */
	GtkWidget              *tab_label;      /* holder for title in tab */
	GTlsCertificateFlags    tls_error;      /* last tls error */
	GtkWidget              *vbx;            /* vertical box (view + entry) */
	GtkWidget              *wv;             /* web view */
};

typedef struct {
	const char             *e;
	enum config_type        t;
	Arg                     v;
} Config;

struct Global {
	int                     clients;
	gboolean                state_lock;
	enum ipc_type           ipc;
	int                     ipc_pipe_fd;
	gchar                  *search_text;
} gl;

struct MainWindow
{
	GtkWidget              *dbx;
	GtkWidget              *nb;
	GtkWidget              *win;
} mw;

/* "normal" functions */
static struct Client *client_create(const gchar *, WebKitWebView *, gboolean,
                                    gboolean);
static void client_destroy(struct Client *);
static gboolean command(struct Client *, const gchar*);
static void create_context_menu(struct Client *, WebKitContextMenu *,
                                WebKitHitTestResult *);
static void die(const char *, ...);
static gboolean ipc_request(GIOChannel *, GIOCondition, gpointer);
static ssize_t ipc_send(char *);
static void ipc_setup(void);
static gboolean key_common(struct Client *, GdkEventKey *);
static gboolean key_entry(struct Client *, GdkEventKey *);
static gboolean key_tab(struct Client *, GdkEvent *);
static gboolean key_web_view(struct Client *, GdkEvent *);
static void load_changed(struct Client *, WebKitLoadEvent);
static void load_configuration(void);
static void load_state(void);
static void load_user_styles(WebKitWebView *);
static void main_window_setup(void);
static void open_external(struct Client *);
static void prepare_download(WebKitDownload *, gchar *);
static void quit(void);
static void render_tls_error(struct Client *, gchar *, GTlsCertificate *,
                             GTlsCertificateFlags);
static void run_user_scripts(WebKitWebView *);
static void save_history(const gchar *);
static void save_state(void);
static void search(struct Client *, enum inpage_search_type);
static void set_default_web_context(void);
static void set_hover_uri(struct Client *, WebKitHitTestResult *);
static void set_javascript_policy(struct Client *, enum javascript_policy);
static void set_window_title(gint);
static void show_web_view(struct Client *);
static void toggle_tls_error_policy(struct Client *);
static gchar *resolve_uri(const gchar *);
static void update_download_button(WebKitDownload *, GtkButton *, gboolean);
static void update_favicon(struct Client *);
static void update_load_progress(struct Client *);
static void update_title(struct Client *);
static void uri_changed(struct Client *);
static void web_view_crashed(struct Client *);

/* callback functions */
static gboolean cb_context_menu(WebKitWebView *, WebKitContextMenu *,
                                GdkEvent *, WebKitHitTestResult *, gpointer);
static void cb_download_cancel(GtkWidget *, gpointer);
static void cb_download_changed_progress(GObject *, GParamSpec *, gpointer);
static void cb_download_finished(GObject *, gpointer);
static gboolean cb_download_prepare(WebKitDownload *, gchar *, gpointer);
static void cb_download_start(WebKitWebView *, WebKitDownload *, gpointer);
static void cb_favicon_changed(GObject *, GParamSpec *, gpointer);
static gboolean cb_entry_hid(GtkWidget *, GdkEvent *, gpointer);
static void cb_entry_icon_hid(GtkEntry *, GtkEntryIconPosition, GdkEvent *,
                              gpointer);
static void cb_notebook_modified(GtkNotebook *, GtkWidget *, guint, gpointer);
static void cb_notebook_switch_page(GtkNotebook *, GtkWidget *, guint,
                                    gpointer);
static gboolean cb_open_external(GSimpleAction *, GVariant *, gpointer);
static void cb_quit(GObject *, gpointer);
static gboolean cb_tab_hid(GtkWidget *, GdkEvent *, gpointer);
static void cb_title_changed(GObject *, GParamSpec *, gpointer);
static void cb_uri_changed(GObject *, GParamSpec *, gpointer);
static void cb_wv_close(GtkWidget *, gpointer);
static gboolean cb_wv_crashed(WebKitWebView *, gpointer);
static WebKitWebView *cb_wv_create(WebKitWebView *, WebKitNavigationAction *,
                                   gpointer);
static gboolean cb_wv_decide_policy(WebKitWebView *, WebKitPolicyDecision *,
                                    WebKitPolicyDecisionType, gpointer);
static gboolean cb_wv_hid(GtkWidget *, GdkEvent *, gpointer);
static void cb_wv_hover(WebKitWebView *, WebKitHitTestResult *, guint,
                        gpointer);
static void cb_wv_load_changed(GObject *, GParamSpec *, gpointer);
static void cb_wv_load_progress_changed(GObject *, GParamSpec *, gpointer);
static void cb_wv_show(WebKitWebView *, gpointer);
static gboolean cb_wv_tls_load_failed(GObject *, gchar *, GTlsCertificate *,
                                      GTlsCertificateFlags, gpointer);


gboolean
cb_context_menu(WebKitWebView *web_view, WebKitContextMenu *context_menu,
                GdkEvent *event, WebKitHitTestResult *hit_test_result,
                gpointer data)
{
	(void)web_view;
	(void)event;

	create_context_menu((struct Client *)data, context_menu, hit_test_result);
	return FALSE;
}

void
cb_download_cancel(GtkWidget *btn, gpointer data)
{
	webkit_download_cancel(WEBKIT_DOWNLOAD(data));
	g_object_unref(WEBKIT_DOWNLOAD(data));
	gtk_widget_destroy(btn);
}

void
cb_download_changed_progress(GObject *obj, GParamSpec *pspec, gpointer data)
{
	(void)pspec;

	update_download_button(WEBKIT_DOWNLOAD(obj), GTK_BUTTON(data), FALSE);
}

void
cb_download_finished(GObject *obj, gpointer data)
{
	if (!GTK_IS_BUTTON(data))
		return;

	update_download_button(WEBKIT_DOWNLOAD(obj), GTK_BUTTON(data), TRUE);
	gtk_button_set_image(
	    GTK_BUTTON(data),
	    gtk_image_new_from_icon_name(ICON_FINISHED, GTK_ICON_SIZE_BUTTON)
	);
}

gboolean
cb_download_prepare(WebKitDownload *download, gchar *suggested_filename,
                    gpointer data)
{
	(void)data;

	prepare_download(download, suggested_filename);
	return FALSE;
}

void
cb_download_start(WebKitWebView *web_view, WebKitDownload *download,
                  gpointer data)
{
	(void)web_view;

	CB(download, "decide-destination", cb_download_prepare, data);
}

void
cb_favicon_changed(GObject *obj, GParamSpec *pspec, gpointer data)
{
	(void)obj;
	(void)pspec;

	update_favicon((struct Client *)data);
}

gboolean
cb_entry_hid(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	(void)widget;

	/* only handle key presses */
	if (event->type == GDK_KEY_PRESS)
		return key_entry((struct Client *)data, (GdkEventKey*)event);

	return FALSE;

}

void
cb_entry_icon_hid(GtkEntry *entry, GtkEntryIconPosition icon_pos,
                  GdkEvent *event, gpointer data)
{
	(void)entry;
	(void)event;

	if (icon_pos == GTK_ENTRY_ICON_PRIMARY) {
		toggle_tls_error_policy((struct Client *)data);
	} else if (icon_pos == GTK_ENTRY_ICON_SECONDARY) {
		set_javascript_policy((struct Client *)data, JSP_TOGGLE);
	}
}

void
cb_notebook_modified(GtkNotebook *notebook, GtkWidget *child, guint idx,
                     gpointer data)
{
	(void)notebook;
	(void)child;
	(void)idx;
	(void)data;

	save_state();
}

void
cb_notebook_switch_page(GtkNotebook *notebook, GtkWidget *child, guint idx,
                        gpointer data)
{
	(void)notebook;
	(void)child;
	(void)data;

	set_window_title(idx);
}

gboolean
cb_open_external(GSimpleAction *action, GVariant *parameter, gpointer data)
{
	(void)action;
	(void)parameter;

	open_external((struct Client *)data);
	return FALSE;
}

void
cb_quit(GObject *obj, gpointer data)
{
	(void)obj;
	(void)data;

	quit();
}

gboolean
cb_tab_hid(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	(void)widget;

	return key_tab((struct Client *)data, event);
}

void
cb_title_changed(GObject *obj, GParamSpec *pspec, gpointer data)
{
	(void)obj;
	(void)pspec;

	update_title((struct Client *)data);
}

void
cb_uri_changed(GObject *obj, GParamSpec *pspec, gpointer data)
{
	(void)obj;
	(void)pspec;

	uri_changed((struct Client *)data);
}

void
cb_wv_close(GtkWidget *widget, gpointer data)
{
	(void)widget;

	client_destroy((struct Client *)data);
}

gboolean
cb_wv_crashed(WebKitWebView *web_view, gpointer data)
{
	(void)web_view;

	web_view_crashed((struct Client *)data);
	return TRUE;
}

WebKitWebView *
cb_wv_create(WebKitWebView *web_view,
             WebKitNavigationAction *navigation_action, gpointer data)
{
	(void)navigation_action;
	(void)data;

	return WEBKIT_WEB_VIEW(client_create(NULL, web_view, FALSE, FALSE)->wv);
}

gboolean
cb_wv_decide_policy(WebKitWebView *web_view, WebKitPolicyDecision *decision,
                    WebKitPolicyDecisionType type, gpointer data)
{
	(void)web_view;
	(void)data;

	if (type != WEBKIT_POLICY_DECISION_TYPE_RESPONSE)
		return FALSE;
	else if (!webkit_response_policy_decision_is_mime_type_supported(
	            WEBKIT_RESPONSE_POLICY_DECISION(decision))
	        )
		webkit_policy_decision_download(decision);
	else
		webkit_policy_decision_use(decision);
	return TRUE;
}

gboolean
cb_wv_hid(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	(void)widget;

	return key_web_view((struct Client *)data, event);
}

void
cb_wv_hover(WebKitWebView *web_view, WebKitHitTestResult *hit_test_result,
            guint modifiers, gpointer data)
{
	(void)web_view;
	(void)modifiers;

	set_hover_uri((struct Client *)data, hit_test_result);
}

void
cb_wv_load_changed(GObject *obj, GParamSpec *pspec, gpointer data)
{
	(void)obj;

	load_changed((struct Client *)data, (WebKitLoadEvent)pspec);
}

void
cb_wv_load_progress_changed(GObject *obj, GParamSpec *pspec, gpointer data)
{
	(void)obj;
	(void)pspec;

	update_load_progress((struct Client *)data);
}

void
cb_wv_show(WebKitWebView *web_view, gpointer data)
{
	(void)web_view;

	show_web_view((struct Client *)data);
}

gboolean
cb_wv_tls_load_failed(GObject *obj, gchar *uri, GTlsCertificate *crt,
                      GTlsCertificateFlags err, gpointer data)
{
	(void)obj;

	render_tls_error((struct Client *)data, uri, crt, err);
	return TRUE;
}

#endif /* !BROWSER_H */
