/**
 * Copyright (C) 2023 Chris Noxz
 * Author(s): Chris Noxz <chris@noxz.tech>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <glib.h>
#include <webkit2/webkit-web-extension.h>

#define     RULE_FILE               "redirect"

GSList      *redirect_patterns      = NULL,
            *redirect_destinations  = NULL;
gboolean    verbose                 = FALSE;


void
redirect_load(void)
{
	GRegex         *r = NULL;                   /* regex rule */
	GError         *e = NULL;                   /* errors */
	GIOChannel     *c = NULL;                   /* reading io channel */
	gchar          *p = NULL,                   /* path to redirect file */
	               *l = NULL,                   /* line buffer */
	                s;                          /* separator holder */
	int             i,
	                j,
	                d;                          /* destination index */

	/* make it possible to debug redirect rules */
	verbose = g_getenv(__NAME_UPPERCASE__"_VERBOSE") != NULL;

	p = g_build_filename(g_get_user_config_dir(), __NAME__, RULE_FILE, NULL);
	if ((c = g_io_channel_new_file(p, "r", NULL)) != NULL) {
		while (g_io_channel_read_line(c, &l, NULL, NULL, NULL)
		       == G_IO_STATUS_NORMAL) {
			g_strstrip(l);
			/* ignore if separator is comment indicator, if not then parse as:
			   /pattern/replacement/ where '/' can be anything but '#' */
			if ((s = l[0]) != '#') {
				for (d = j = i = 0; l[i] && j <= 2; i++) {
					if (d == 0 && j == 1)
						d = i;
					l[i] = l[i + 1];
					if (l[i] == s) {
						l[i] = '\0';
						j++;
					}
				}

				if (j == 2) { /* make sure to pass 2 null-terminated strings */
					r = g_regex_new(
					    l,
					    G_REGEX_CASELESS | G_REGEX_OPTIMIZE,
					    G_REGEX_MATCH_PARTIAL,
					    &e
					);
					if (e != NULL) {
						fprintf(
						    stderr,
						    __NAME__": Could not compile regex: %s\n",
						    l
						);
						g_error_free(e);
						e = NULL;
					} else {
						redirect_patterns = g_slist_append(
						    redirect_patterns, r
						);
						redirect_destinations = g_slist_append(
						    redirect_destinations, g_strdup(l + d)
						);
					}
				}
			}
			g_free(l);
		}
		g_io_channel_shutdown(c, FALSE, NULL);
	}
	g_free(p);
}

gboolean
cb_web_page_send_request(WebKitWebPage *web_page, WebKitURIRequest *request,
                         WebKitURIResponse *redirected_response, gpointer data)
{
	GSList         *p = redirect_patterns,      /* pattern holder */
	               *d = redirect_destinations;  /* destination holder */
	const gchar    *i;                          /* incoming uri */
	gchar          *o = NULL;                   /* outgoing uri */

	(void)web_page;
	(void)redirected_response;
	(void)data;

	i = webkit_uri_request_get_uri(request);

	/* go through every rule and apply accordingly */
	while (p && d) {
		/* find matching rule for incoming uri */
		if (g_regex_match((GRegex *)(p->data), i, 0, NULL)) {
			/* check if destination exists for rule */
			if (((gchar*)d->data)[0])
				o = g_regex_replace(
				    (GRegex *)(p->data),
				    i,
				    -1,
				    0,
				    ((gchar*)d->data),
				    G_REGEX_MATCH_DEFAULT,
				    NULL
				);
			if (verbose)
				fprintf(stderr,
				    "redirect: %s %s -> %s\n",
				    g_regex_get_pattern(p->data),
				    i,
				    o ? o : "[blocked]"
				);
			/* if outgoing uri was constructed, then redirect... */
			if (o) {
				webkit_uri_request_set_uri(request, o);
				g_free(o);
				return FALSE;
			/* ...else, block */
			} else {
				return TRUE;
			}
		}
		p = g_slist_next(p);
		d = g_slist_next(d);
	}

	return FALSE;
}

void
cb_web_page_created(WebKitWebExtension *extension, WebKitWebPage *web_page,
                    gpointer data)
{
	(void)extension;
	(void)data;

	g_signal_connect_object(
	    web_page, "send-request", G_CALLBACK(cb_web_page_send_request), NULL, 0
	);
}

G_MODULE_EXPORT void
webkit_web_extension_initialize(WebKitWebExtension *extension)
{
	redirect_load();
	g_signal_connect(
	    extension, "page-created", G_CALLBACK(cb_web_page_created), NULL
	);
}
